@@ -284,6 +284,7 @@ void increase_comb_counter(size_t ix_arr[], size_t st, size_t end, size_t n,
         }
 }
 
+/* Note to self: don't try merge this into a template with the one above, as the other one has 'restrict' qualifier */
 void increase_comb_counter(size_t ix_arr[], size_t st, size_t end, size_t n,
                            double counter[], std::unordered_map<size_t, double> &weights, double exp_remainder)
 {
@@ -312,6 +313,54 @@ void increase_comb_counter(size_t ix_arr[], size_t st, size_t end, size_t n,
         }
 }
 
+void increase_comb_counter_in_groups(size_t ix_arr[], size_t st, size_t end, size_t split_ix, size_t n,
+                                     double counter[], double exp_remainder)
+{
+    size_t n_group = 0;
+    for (size_t ix = st; ix <= end; ix++)
+        if (ix_arr[ix] < split_ix)
+            n_group++;
+        else
+            break;
+
+    n = n - split_ix;
+
+    if (exp_remainder <= 1)
+        for (size_t ix1 = st; ix1 < st + n_group; ix1++)
+            for (size_t ix2 = st + n_group; ix2 <= end; ix2++)
+                counter[ix_arr[ix1] * n + ix_arr[ix2] - split_ix]++;
+    else
+        for (size_t ix1 = st; ix1 < st + n_group; ix1++)
+            for (size_t ix2 = st + n_group; ix2 <= end; ix2++)
+                counter[ix_arr[ix1] * n + ix_arr[ix2] - split_ix] += exp_remainder;
+}
+
+void increase_comb_counter_in_groups(size_t ix_arr[], size_t st, size_t end, size_t split_ix, size_t n,
+                                     double *restrict counter, double *restrict weights, double exp_remainder)
+{
+    size_t n_group = 0;
+    for (size_t ix = st; ix <= end; ix++)
+        if (ix_arr[ix] < split_ix)
+            n_group++;
+        else
+            break;
+
+    n = n - split_ix;
+
+    if (exp_remainder <= 1)
+        for (size_t ix1 = st; ix1 < st + n_group; ix1++)
+            for (size_t ix2 = st + n_group; ix2 <= end; ix2++)
+                counter[ix_arr[ix1] * n + ix_arr[ix2] - split_ix]
+                    +=
+                weights[ix_arr[ix1]] * weights[ix_arr[ix2]];
+    else
+        for (size_t ix1 = st; ix1 < st + n_group; ix1++)
+            for (size_t ix2 = st + n_group; ix2 <= end; ix2++)
+                counter[ix_arr[ix1] * n + ix_arr[ix2] - split_ix]
+                    +=
+                weights[ix_arr[ix1]] * weights[ix_arr[ix2]] * exp_remainder;
+}
+
 void tmat_to_dense(double *restrict tmat, double *restrict dmat, size_t n, bool diag_to_one)
 {
     size_t ncomb = (n * (n - 1)) / 2;
@@ -438,7 +487,8 @@ void sample_random_rows(std::vector<size_t> &ix_arr, size_t nrows, bool with_rep
             for (size_t lev = 0; lev < log2_n; lev++)
             {
                 curr_ix = ix_parent(curr_ix);
-                btree_weights[curr_ix] -= sample_weights[ix];
+                btree_weights[curr_ix] =   btree_weights[ix_child(curr_ix)]
+                                         + btree_weights[ix_child(curr_ix) + 1];
             }
         }
     }
@@ -572,7 +622,7 @@ void weighted_shuffle(size_t *restrict outp, size_t n, double *restrict weights,
         curr_subrange = buffer_arr[0];
         for (size_t lev = 0; lev < tree_levels; lev++)
         {
-            rnd_subrange = std::uniform_real_distribution<double>(0, curr_subrange)(rnd_generator);
+            rnd_subrange = std::uniform_real_distribution<double>(0., curr_subrange)(rnd_generator);
             w_left = buffer_arr[ix_child(curr_ix)];
             curr_ix = ix_child(curr_ix) + (rnd_subrange >= w_left);
             curr_subrange = buffer_arr[curr_ix];
@@ -586,7 +636,8 @@ void weighted_shuffle(size_t *restrict outp, size_t n, double *restrict weights,
         for (size_t lev = 0; lev < tree_levels; lev++)
         {
             curr_ix = ix_parent(curr_ix);
-            buffer_arr[curr_ix] -= weights[outp[el]];
+            buffer_arr[curr_ix] =   buffer_arr[ix_child(curr_ix)]
+                                  + buffer_arr[ix_child(curr_ix) + 1];
         }
     }
 
@@ -1499,3 +1550,25 @@ void todense(size_t ix_arr[], size_t st, size_t end,
         }
     }
 }
+
+/* Function to handle interrupt signals */
+void set_interrup_global_variable(int s)
+{
+    fprintf(stderr, "Error: procedure was interrupted\n");
+    #pragma omp critical
+    {
+        interrupt_switch = true;
+    }
+}
+
+/* Return the #def'd constants from standard header. This is in order to determine if the return
+   value from the 'fit_model' function is a success or failure within Cython, which does not
+   allow importing #def'd macro values. */
+int return_EXIT_SUCCESS()
+{
+    return EXIT_SUCCESS;
+}
+int return_EXIT_FAILURE()
+{
+    return EXIT_FAILURE;
+}
